;; Generated package description from ellama.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "ellama" "1.31.0.0.20260720.1" "Tool for interacting with LLMs" '((emacs "28.1") (llm "0.31.1") (plz "0.8") (transient "0.7") (compat "29.1") (yaml "1.2.3")) :commit "5d32a48a8d1797fec776ba56279dc99dd58f399a" :authors '(("Sergey Kostyaev" . "sskostyaev@gmail.com")) :maintainer '("Sergey Kostyaev" . "sskostyaev@gmail.com") :keywords '("help" "local" "tools") :url "http://github.com/s-kostyaev/ellama")
